package orig.plotter;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.Line2D;

import javax.swing.JFrame;
import javax.swing.JPanel;

/**
 * Simulator for a multi-pen plotter.
 * 
 * @author evensen
 * 
 */
public class ColorPenPlotter implements Plotter {
	private boolean penDown = false;
	private PenColor penColor = PenColor.BLACK;
	private final Graphics2D bufferGraphics;
	private final long drawingSpeed;
	private final long transportSpeed;
	private int x;
	private int y;
	private final Image imgBuffer;

	/**
	 * Creates a new window with a drawing area.
	 * 
	 * The pen-head will be lifted and at a random position from the start.
	 * 
	 * @param width
	 *            The width of the paper/window in pixels.
	 * @param height
	 *            The height of the paper/window in pixels.
	 * @param drawingSpeed
	 *            The moving speed for the pen-head while drawing.
	 * @param transportSpeed
	 *            The moving speed for the pen-head when the pen is lifted.
	 */
	public ColorPenPlotter(final int width, final int height,
			final int drawingSpeed,
			final int transportSpeed, final int fps) {
		JFrame f = new JFrame("PlotterSimulator");
		final JPanel p = new JPanel(true);
		p.setMinimumSize(new Dimension(width, height));
		p.setPreferredSize(p.getMinimumSize());
		p.setVisible(true);
		p.setOpaque(true);

		f.add(p);
		f.pack();
		f.setVisible(true);

		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		this.imgBuffer = p.createImage(width, height);
		this.bufferGraphics = (Graphics2D) this.imgBuffer.getGraphics();
		this.bufferGraphics.setColor(Color.WHITE);
		this.bufferGraphics.fillRect(0, 0, width, height);
		this.bufferGraphics.setColor(Color.BLACK);
		this.bufferGraphics.setRenderingHint(RenderingHints.KEY_RENDERING,
				RenderingHints.VALUE_RENDER_QUALITY);
		this.bufferGraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		this.bufferGraphics.setStroke(new BasicStroke(0.5f,
				BasicStroke.CAP_ROUND, BasicStroke.JOIN_MITER));

		p.setBackground(Color.WHITE);
		this.x = (int) (Math.random() * width);
		this.y = (int) (Math.random() * height);
		this.drawingSpeed = drawingSpeed;
		this.transportSpeed = transportSpeed;

		Thread redrawer = new Thread() {
			@Override
			public void run() {
				super.run();
				Graphics2D g = (Graphics2D) p.getGraphics();
				g.setColor(Color.white);
				g.fillRect(0, 0, p.getWidth(), p.getHeight());
				g.drawImage(ColorPenPlotter.this.imgBuffer, 0, 0, p);
				int delay = 1000000000 / fps;
				while (true) {
					g.setColor(Color.white);
					// g.fillRect(0, 0, p.getWidth(), p.getHeight());
					g.drawImage(ColorPenPlotter.this.imgBuffer, 0, 0, p);
					long previousCallTime = System.nanoTime();

					while (System.nanoTime() - previousCallTime < delay) {
						synchronized (ColorPenPlotter.this.imgBuffer) {
							try {
								ColorPenPlotter.this.imgBuffer.wait();
							} catch (InterruptedException e) {
								Thread.currentThread().interrupt();
							}
						}
					}

				}
			}
		};

		redrawer.start();
	}

	/**
	 * Fetches and changes to a different coloured pen.
	 * 
	 * Lifts the pen, moves the pen-head to the pen storage and stops there.
	 * 
	 * @param color
	 */
	@Override
	public void setPenColor(final PenColor color) {
		if (color != this.penColor) {
			penDown(false);
			moveTo(color.ordinal() * 20, 0);
			this.penColor = color;
			this.bufferGraphics.setColor(color.getColor());
		}
	}

	/**
	 * Lifts the pen and moves to the origin of the paper.
	 */
	@Override
	public void home() {
		penDown(false);
		moveTo(0, 0);
	}

	/**
	 * Lowers or lifts the pen.
	 * 
	 * @param v
	 *            <code>true</code> lowers the pen, <code>false</code> lifts it.
	 */
	@Override
	public void penDown(final boolean v) {
		if (v && !this.penDown) {
			this.bufferGraphics.drawOval(this.x, this.y, 1, 1);
		}
		this.penDown = v;
	}

	private void flashOn(final int xc, final int yc) {
		this.bufferGraphics.setXORMode(Color.GREEN);
		this.bufferGraphics.fillOval(xc - 1, yc - 1, 3, 3);
		synchronized (this.imgBuffer) {
			this.imgBuffer.notifyAll();
		}
	}

	private void flashOff(final int xc, final int yc) {
		this.bufferGraphics.fillOval(xc - 1, yc - 1, 3, 3);
		this.bufferGraphics.setPaintMode();
		this.bufferGraphics.setColor(this.penColor.getColor());
		synchronized (this.imgBuffer) {
			this.imgBuffer.notifyAll();
		}
	}

	/**
	 * Moves (and possibly draws) from the current position to the coordinate
	 * given as <x1, y1>.
	 * 
	 * @param x1
	 *            ,y1 the coordinate to move to.
	 */
	@Override
	public void moveTo(final int x1, final int y1) {
		int deltaX = Math.abs(x1 - this.x);
		int deltaY = Math.abs(y1 - this.y);
		boolean steep = deltaY > deltaX;
		long slowness =
				1000 - (this.penDown ? this.drawingSpeed : this.transportSpeed);

		if (deltaX > 0 || deltaY > 0) {
			if (steep) { // Loop over y
				double xStep = (x1 - this.x) / (double) deltaY;
				int yDir = (int) Math.signum(y1 - this.y);
				for (int yc = this.y, s = 0; yc != y1; yc += yDir) {
					int xc = (int) (this.x + xStep * s++);
					flashOn(xc, yc);
					if (slowness > 0) {
						try {
							Thread.sleep(slowness);
						} catch (InterruptedException e) {
							Thread.currentThread().interrupt();
						}
					}
					flashOff(xc, yc);

					if (this.penDown) { // Paint a pixel
						this.bufferGraphics.drawLine(xc, yc, xc, yc);

					}
				}
			} else { // Loop over x
				double yStep = (y1 - this.y) / (double) deltaX;
				int xDir = (int) Math.signum(x1 - this.x);
				for (int xc = this.x, s = 0; xc != x1; xc += xDir) {
					int yc = (int) (this.y + yStep * s++);
					flashOn(xc, yc);
					if (slowness > 0) {
						try {
							Thread.sleep(slowness);
						} catch (InterruptedException e) {
							Thread.currentThread().interrupt();
						}
					}
					flashOff(xc, yc);

					if (this.penDown) { // Paint a pixel
						this.bufferGraphics.drawLine(xc, yc, xc, yc);
					}
				}
			}
		}
		if (this.penDown) { // Paint a pixel
			this.bufferGraphics.drawLine(x1, y1, x1, y1);
			Shape l = new Line2D.Double(this.x, this.y, x1, y1);
			this.bufferGraphics.draw(l);
		}
		this.x = x1;
		this.y = y1;
		try {
			Thread.sleep(1);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}
	}

}
